/*
 *  Copyright (C) 2000-2002 Constantin Kaplinsky.  All Rights Reserved.
 *  Copyright (C) 2000 Tridia Corporation.  All Rights Reserved.
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * rfbproto.c - functions to deal with client side of RFB protocol.
 */

#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <vncviewer.h>
#include <vncauth.h>
#include <zlib.h>
#include <jpeglib.h>

static void InitCapabilities(void);
static Bool SetupTunneling(void);
static int ReadSecurityType(void);
static int SelectSecurityType(void);
static Bool PerformAuthenticationTight(void);
static Bool AuthenticateVNC(void);
static Bool AuthenticateUnixLogin(void);
static Bool ReadInteractionCaps(void);
static Bool ReadCapabilityList(CapsContainer *caps, int count);

static Bool HandleRRE8(int rx, int ry, int rw, int rh);
static Bool HandleRRE16(int rx, int ry, int rw, int rh);
static Bool HandleRRE32(int rx, int ry, int rw, int rh);
static Bool HandleCoRRE8(int rx, int ry, int rw, int rh);
static Bool HandleCoRRE16(int rx, int ry, int rw, int rh);
static Bool HandleCoRRE32(int rx, int ry, int rw, int rh);
static Bool HandleHextile8(int rx, int ry, int rw, int rh);
static Bool HandleHextile16(int rx, int ry, int rw, int rh);
static Bool HandleHextile32(int rx, int ry, int rw, int rh);
static Bool HandleZlib8(int rx, int ry, int rw, int rh);
static Bool HandleZlib16(int rx, int ry, int rw, int rh);
static Bool HandleZlib32(int rx, int ry, int rw, int rh);
static Bool HandleTight8(int rx, int ry, int rw, int rh);
static Bool HandleTight16(int rx, int ry, int rw, int rh);
static Bool HandleTight32(int rx, int ry, int rw, int rh);

/* runge add zrle */
static Bool HandleZRLE8(int rx, int ry, int rw, int rh);
static Bool HandleZRLE16(int rx, int ry, int rw, int rh);
static Bool HandleZRLE24(int rx, int ry, int rw, int rh);
static Bool HandleZRLE24Up(int rx, int ry, int rw, int rh);
static Bool HandleZRLE24Down(int rx, int ry, int rw, int rh);
static Bool HandleZRLE32(int rx, int ry, int rw, int rh);

typedef struct {
    unsigned long length;
} rfbZRLEHeader;

long zywrle;

#define sz_rfbZRLEHeader 4

#define rfbZRLETileWidth 64
#define rfbZRLETileHeight 64

static void ReadConnFailedReason(void);
static long ReadCompactLen (void);

static void JpegInitSource(j_decompress_ptr cinfo);
static boolean JpegFillInputBuffer(j_decompress_ptr cinfo);
static void JpegSkipInputData(j_decompress_ptr cinfo, long num_bytes);
static void JpegTermSource(j_decompress_ptr cinfo);
static void JpegSetSrcManager(j_decompress_ptr cinfo, CARD8 *compressedData,
                              int compressedLen);


int sent_FBU = 0;
int skip_XtUpdate = 0;
int skip_XtUpdateAll = 0;

int rfbsock;
char *desktopName;
rfbPixelFormat myFormat;
rfbServerInitMsg si;
char *serverCutText = NULL;
Bool newServerCutText = False;

int endianTest = 1;

static Bool tightVncProtocol = False;
static CapsContainer *tunnelCaps;    /* known tunneling/encryption methods */
static CapsContainer *authCaps;	     /* known authentication schemes       */
static CapsContainer *serverMsgCaps; /* known non-standard server messages */
static CapsContainer *clientMsgCaps; /* known non-standard client messages */
static CapsContainer *encodingCaps;  /* known encodings besides Raw        */


/* Note that the CoRRE encoding uses this buffer and assumes it is big enough
   to hold 255 * 255 * 32 bits -> 260100 bytes.  640*480 = 307200 bytes.
   Hextile also assumes it is big enough to hold 16 * 16 * 32 bits.
   Tight encoding assumes BUFFER_SIZE is at least 16384 bytes. */

#define BUFFER_SIZE (640*480)
static char buffer[BUFFER_SIZE];


/* The zlib encoding requires expansion/decompression/deflation of the
   compressed data in the "buffer" above into another, result buffer.
   However, the size of the result buffer can be determined precisely
   based on the bitsPerPixel, height and width of the rectangle.  We
   allocate this buffer one time to be the full size of the buffer. */

static int raw_buffer_size = -1;
static char *raw_buffer;

static z_stream decompStream;
static Bool decompStreamInited = False;


/*
 * Variables for the ``tight'' encoding implementation.
 */

/* Separate buffer for compressed data. */
#define ZLIB_BUFFER_SIZE 512
static char zlib_buffer[ZLIB_BUFFER_SIZE];

/* Four independent compression streams for zlib library. */
static z_stream zlibStream[4];
static Bool zlibStreamActive[4] = {
  False, False, False, False
};

/* Filter stuff. Should be initialized by filter initialization code. */
static Bool cutZeros;
static int rectWidth, rectColors;
static char tightPalette[256*4];
static CARD8 tightPrevRow[2048*3*sizeof(CARD16)];

/* JPEG decoder state. */
static Bool jpegError;


/*
 * InitCapabilities.
 */

static void
InitCapabilities(void)
{
  tunnelCaps    = CapsNewContainer();
  authCaps      = CapsNewContainer();
  serverMsgCaps = CapsNewContainer();
  clientMsgCaps = CapsNewContainer();
  encodingCaps  = CapsNewContainer();

  /* Supported authentication methods */
  CapsAdd(authCaps, rfbAuthVNC, rfbStandardVendor, sig_rfbAuthVNC,
	  "Standard VNC password authentication");
  CapsAdd(authCaps, rfbAuthUnixLogin, rfbTightVncVendor, sig_rfbAuthUnixLogin,
	  "Login-style Unix authentication");

  /* Supported encoding types */
  CapsAdd(encodingCaps, rfbEncodingCopyRect, rfbStandardVendor,
	  sig_rfbEncodingCopyRect, "Standard CopyRect encoding");
  CapsAdd(encodingCaps, rfbEncodingRRE, rfbStandardVendor,
	  sig_rfbEncodingRRE, "Standard RRE encoding");
  CapsAdd(encodingCaps, rfbEncodingCoRRE, rfbStandardVendor,
	  sig_rfbEncodingCoRRE, "Standard CoRRE encoding");
  CapsAdd(encodingCaps, rfbEncodingHextile, rfbStandardVendor,
	  sig_rfbEncodingHextile, "Standard Hextile encoding");
  CapsAdd(encodingCaps, rfbEncodingZlib, rfbTridiaVncVendor,
	  sig_rfbEncodingZlib, "Zlib encoding from TridiaVNC");
  CapsAdd(encodingCaps, rfbEncodingTight, rfbTightVncVendor,
	  sig_rfbEncodingTight, "Tight encoding by Constantin Kaplinsky");

  /* Supported "fake" encoding types */
  CapsAdd(encodingCaps, rfbEncodingCompressLevel0, rfbTightVncVendor,
	  sig_rfbEncodingCompressLevel0, "Compression level");
  CapsAdd(encodingCaps, rfbEncodingQualityLevel0, rfbTightVncVendor,
	  sig_rfbEncodingQualityLevel0, "JPEG quality level");
  CapsAdd(encodingCaps, rfbEncodingXCursor, rfbTightVncVendor,
	  sig_rfbEncodingXCursor, "X-style cursor shape update");
  CapsAdd(encodingCaps, rfbEncodingRichCursor, rfbTightVncVendor,
	  sig_rfbEncodingRichCursor, "Rich-color cursor shape update");
  CapsAdd(encodingCaps, rfbEncodingPointerPos, rfbTightVncVendor,
	  sig_rfbEncodingPointerPos, "Pointer position update");
  CapsAdd(encodingCaps, rfbEncodingLastRect, rfbTightVncVendor,
	  sig_rfbEncodingLastRect, "LastRect protocol extension");

  CapsAdd(encodingCaps, rfbEncodingNewFBSize, rfbTightVncVendor,
	  sig_rfbEncodingNewFBSize, "New FB size protocol extension");
}


/*
 * ConnectToRFBServer.
 */

Bool
ConnectToRFBServer(const char *hostname, int port)
{
	unsigned int host;

	if (!StringToIPAddr(hostname, &host)) {
		fprintf(stderr,"Couldn't convert '%s' to host address\n", hostname);
		return False;
	}

	rfbsock = ConnectToTcpAddr(host, port);

	if (rfbsock < 0) {
		fprintf(stderr,"Unable to connect to VNC server\n");
		return False;
	}

	return SetNonBlocking(rfbsock);
}


/*
 * InitialiseRFBConnection.
 */

Bool
InitialiseRFBConnection(void)
{
	rfbProtocolVersionMsg pv;
	int server_major, server_minor;
	int viewer_major, viewer_minor;
	rfbClientInitMsg ci;
	int secType;

	/* if the connection is immediately closed, don't report anything, so
	   that pmw's monitor can make test connections */

	if (listenSpecified) {
		errorMessageOnReadFailure = False;
	}

	if (!ReadFromRFBServer(pv, sz_rfbProtocolVersionMsg)) {
		return False;
	}

	errorMessageOnReadFailure = True;

	pv[sz_rfbProtocolVersionMsg] = 0;

	if (sscanf(pv, rfbProtocolVersionFormat, &server_major, &server_minor) != 2) {
		fprintf(stderr,"Not a valid VNC server\n");
		return False;
	}

	viewer_major = rfbProtocolMajorVersion;
	if (server_major == 3 && server_minor >= rfbProtocolMinorVersion) {
		/* the server supports at least the standard protocol 3.7 */
		viewer_minor = rfbProtocolMinorVersion;
	} else {
		/* any other server version, request the standard 3.3 */
		viewer_minor = rfbProtocolFallbackMinorVersion;
	}

	fprintf(stderr, "\nConnected to RFB server, using protocol version %d.%d\n", viewer_major, viewer_minor);

	sprintf(pv, rfbProtocolVersionFormat, viewer_major, viewer_minor);

	if (!WriteExact(rfbsock, pv, sz_rfbProtocolVersionMsg)) {
		return False;
	}

	/* Read or select the security type. */
	if (viewer_minor == rfbProtocolMinorVersion) {
		secType = SelectSecurityType();
	} else {
		secType = ReadSecurityType();
	}
	if (secType == rfbSecTypeInvalid) {
		return False;
	}

	switch (secType) {
		case rfbSecTypeNone:
			fprintf(stderr, "No authentication needed\n\n");
			break;
		case rfbSecTypeVncAuth:
			if (!AuthenticateVNC()) {
				return False;
			}
			break;
		case rfbSecTypeTight:
			tightVncProtocol = True;
			InitCapabilities();
			if (!SetupTunneling()) {
				return False;
			}
			if (!PerformAuthenticationTight()) {
				return False;
			}
			break;
		default:                      /* should never happen */
			fprintf(stderr, "Internal error: Invalid security type\n");
			return False;
	}

	ci.shared = (appData.shareDesktop ? 1 : 0);

	if (!WriteExact(rfbsock, (char *)&ci, sz_rfbClientInitMsg)) {
		return False;
	}

	if (!ReadFromRFBServer((char *)&si, sz_rfbServerInitMsg)) {
		return False;
	}

	si.framebufferWidth = Swap16IfLE(si.framebufferWidth);
	si.framebufferHeight = Swap16IfLE(si.framebufferHeight);
	si.format.redMax = Swap16IfLE(si.format.redMax);
	si.format.greenMax = Swap16IfLE(si.format.greenMax);
	si.format.blueMax = Swap16IfLE(si.format.blueMax);
	si.nameLength = Swap32IfLE(si.nameLength);

	/* FIXME: Check arguments to malloc() calls. */
	desktopName = malloc(si.nameLength + 1);
	if (!desktopName) {
		fprintf(stderr, "Error allocating memory for desktop name, %lu bytes\n",
		    (unsigned long)si.nameLength);
		return False;
	}

	if (!ReadFromRFBServer(desktopName, si.nameLength)) {
		return False;
	}

	desktopName[si.nameLength] = 0;

	fprintf(stderr,"Desktop name \"%s\"\n\n", desktopName);

	fprintf(stderr,"VNC server default format:\n");
	PrintPixelFormat(&si.format);

	if (tightVncProtocol) {
		/* Read interaction capabilities (protocol 3.7t) */
		if (!ReadInteractionCaps()) {
			return False;
		}
	}

	return True;
}


/*
 * Read security type from the server (protocol 3.3)
 */

static int
ReadSecurityType(void)
{
	CARD32 secType;

	/* Read the security type */
	if (!ReadFromRFBServer((char *)&secType, sizeof(secType))) {
		return rfbSecTypeInvalid;
	}

	secType = Swap32IfLE(secType);

	if (secType == rfbSecTypeInvalid) {
		ReadConnFailedReason();
		return rfbSecTypeInvalid;
	}

	if (secType != rfbSecTypeNone && secType != rfbSecTypeVncAuth) {
		fprintf(stderr, "Unknown security type from RFB server: %d\n",
		    (int)secType);
		return rfbSecTypeInvalid;
	}

	return (int)secType;
}


/*
 * Select security type from the server's list (protocol 3.7)
 */

static int
SelectSecurityType(void)
{
	CARD8 nSecTypes;
	char *secTypeNames[] = {"None", "VncAuth"};
	CARD8 knownSecTypes[] = {rfbSecTypeNone, rfbSecTypeVncAuth};
	int nKnownSecTypes = sizeof(knownSecTypes);
	CARD8 *secTypes;
	CARD8 secType = rfbSecTypeInvalid;
	int i, j;

	/* Read the list of secutiry types. */
	if (!ReadFromRFBServer((char *)&nSecTypes, sizeof(nSecTypes))) {
		return rfbSecTypeInvalid;
	}

	if (nSecTypes == 0) {
		ReadConnFailedReason();
		return rfbSecTypeInvalid;
	}

	secTypes = malloc(nSecTypes);
	if (!ReadFromRFBServer((char *)secTypes, nSecTypes)) {
		return rfbSecTypeInvalid;
	}

	/* Find out if the server supports TightVNC protocol extensions */
	for (j = 0; j < (int)nSecTypes; j++) {
		if (secTypes[j] == rfbSecTypeTight) {
			free(secTypes);
			secType = rfbSecTypeTight;
			if (!WriteExact(rfbsock, (char *)&secType, sizeof(secType))) {
				return rfbSecTypeInvalid;
			}
			fprintf(stderr, "Enabling TightVNC protocol extensions\n");
			return rfbSecTypeTight;
		}
	}

	/* Find first supported security type */
	for (j = 0; j < (int)nSecTypes; j++) {
		for (i = 0; i < nKnownSecTypes; i++) {
			if (secTypes[j] == knownSecTypes[i]) {
				secType = secTypes[j];
				if (!WriteExact(rfbsock, (char *)&secType, sizeof(secType))) {
				  free(secTypes);
				  return rfbSecTypeInvalid;
				}
				break;
				}
		}
		if (secType != rfbSecTypeInvalid) {
			break;
		}
	}

	free(secTypes);

	if (secType == rfbSecTypeInvalid) {
		fprintf(stderr, "Server did not offer supported security type\n");
	}

	return (int)secType;
}


/*
 * Setup tunneling (protocol version 3.7t).
 */

static Bool
SetupTunneling(void)
{
  rfbTunnelingCapsMsg caps;
  CARD32 tunnelType;

  /* In the protocol version 3.7t, the server informs us about
     supported tunneling methods. Here we read this information. */

  if (!ReadFromRFBServer((char *)&caps, sz_rfbTunnelingCapsMsg))
    return False;

  caps.nTunnelTypes = Swap32IfLE(caps.nTunnelTypes);

  if (caps.nTunnelTypes) {
    if (!ReadCapabilityList(tunnelCaps, caps.nTunnelTypes))
      return False;

    /* We cannot do tunneling anyway yet. */
    tunnelType = Swap32IfLE(rfbNoTunneling);
    if (!WriteExact(rfbsock, (char *)&tunnelType, sizeof(tunnelType)))
      return False;
  }

  return True;
}

static char *restart_session_pw = NULL;
static int restart_session_len = 0;


/*
 * Negotiate authentication scheme (protocol version 3.7t)
 */

static Bool
PerformAuthenticationTight(void)
{
	rfbAuthenticationCapsMsg caps;
	CARD32 authScheme;
	int i;

	/* In the protocol version 3.7t, the server informs us about supported
	authentication schemes. Here we read this information. */

	if (!ReadFromRFBServer((char *)&caps, sz_rfbAuthenticationCapsMsg)) {
		return False;
	}

	caps.nAuthTypes = Swap32IfLE(caps.nAuthTypes);

	if (!caps.nAuthTypes) {
		fprintf(stderr, "No authentication needed\n\n");
		return True;
	}

	if (!ReadCapabilityList(authCaps, caps.nAuthTypes)) {
		return False;
	}

	/* Prefer Unix login authentication if a user name was given. */
	if (appData.userLogin && CapsIsEnabled(authCaps, rfbAuthUnixLogin)) {
		authScheme = Swap32IfLE(rfbAuthUnixLogin);
		if (!WriteExact(rfbsock, (char *)&authScheme, sizeof(authScheme))) {
			return False;
		}
		return AuthenticateUnixLogin();
	}

	/* Otherwise, try server's preferred authentication scheme. */
	for (i = 0; i < CapsNumEnabled(authCaps); i++) {
		authScheme = CapsGetByOrder(authCaps, i);
		if (authScheme != rfbAuthUnixLogin && authScheme != rfbAuthVNC) {
			continue;                 /* unknown scheme - cannot use it */
		}
		authScheme = Swap32IfLE(authScheme);
		if (!WriteExact(rfbsock, (char *)&authScheme, sizeof(authScheme))) {
			return False;
		}
		authScheme = Swap32IfLE(authScheme); /* convert it back */
		if (authScheme == rfbAuthUnixLogin) {
			return AuthenticateUnixLogin();
		} else if (authScheme == rfbAuthVNC) {
			return AuthenticateVNC();
		} else {
			/* Should never happen. */
			fprintf(stderr, "Assertion failed: unknown authentication scheme\n");
			return False;
		}
	}

	fprintf(stderr, "No suitable authentication schemes offered by server\n");
	return False;
}


/*
 * Standard VNC authentication.
 */

static Bool
AuthenticateVNC(void)
{
	CARD32 authScheme, authResult;
	CARD8 challenge[CHALLENGESIZE];
	char *passwd = NULL;
	char  buffer[64];
	char* cstatus;
	int   len;
	int restart = 0;

	fprintf(stderr, "\nPerforming standard VNC authentication\n");

	if (!ReadFromRFBServer((char *)challenge, CHALLENGESIZE)) {
		return False;
	}

	if (restart_session_pw != NULL) {
		passwd = restart_session_pw;	
		restart_session_pw = NULL;
		restart = 1;
	} else if (appData.passwordFile) {
		passwd = vncDecryptPasswdFromFile(appData.passwordFile);
		if (!passwd) {
			fprintf(stderr, "Cannot read valid password from file \"%s\"\n",
			      appData.passwordFile);
			return False;
		}
	} else if (appData.autoPass) {
		passwd = buffer;
		cstatus = fgets(buffer, sizeof buffer, stdin);
		if (cstatus == NULL) {
			buffer[0] = '\0';
		} else {
			len = strlen(buffer);
			if (len > 0 && buffer[len - 1] == '\n') {
				buffer[len - 1] = '\0';
			}
		}
	} else if (appData.passwordDialog) {
		passwd = DoPasswordDialog();
	} else {
		passwd = getpass("VNC Password: ");
	}

	if (restart) {
#define EN0 0
#define DE1 1
		unsigned char s_fixedkey[8] = {23,82,107,6,35,78,88,7};
		deskey(s_fixedkey, DE1);
		des(passwd, passwd);
	} else {
		if (!passwd || strlen(passwd) == 0) {
			fprintf(stderr, "Reading password failed\n\n");
			return False;
		}
		if (strlen(passwd) > 8) {
			passwd[8] = '\0';
		}
	}
	vncEncryptBytes(challenge, passwd);
	


//	/* Lose the password from memory */
//	memset(passwd, '\0', strlen(passwd));

	if (!WriteExact(rfbsock, (char *)challenge, CHALLENGESIZE)) {
		return False;
	}

	if (!ReadFromRFBServer((char *)&authResult, 4)) {
		return False;
	}

	authResult = Swap32IfLE(authResult);

	switch (authResult) {
	case rfbVncAuthOK:
		fprintf(stderr, "VNC authentication succeeded\n\n");
		break;
	case rfbVncAuthFailed:
		fprintf(stderr, "VNC authentication failed\n\n");
		return False;
	case rfbVncAuthTooMany:
		fprintf(stderr, "VNC authentication failed - too many tries\n\n");
		return False;
	default:
		fprintf(stderr, "Unknown VNC authentication result: %d\n\n",
		    (int)authResult);
		return False;
	}

	return True;
}

/*
 * Unix login-style authentication.
 */

static Bool
AuthenticateUnixLogin(void)
{
	CARD32 loginLen, passwdLen, authResult;
	char *login;
	char *passwd;
	struct passwd *ps;

	fprintf(stderr, "\nPerforming Unix login-style authentication\n");

	if (appData.userLogin) {
		login = appData.userLogin;
	} else {
		ps = getpwuid(getuid());
		login = ps->pw_name;
	}

	fprintf(stderr, "Using user name \"%s\"\n", login);

	if (appData.passwordDialog) {
		passwd = DoPasswordDialog();
	} else {
		passwd = getpass("VNC Password: ");
	}
	if (!passwd || strlen(passwd) == 0) {
		fprintf(stderr, "Reading password failed\n");
		return False;
	}

	loginLen = Swap32IfLE((CARD32)strlen(login));
	passwdLen = Swap32IfLE((CARD32)strlen(passwd));

	if (!WriteExact(rfbsock, (char *)&loginLen, sizeof(loginLen)) ||
	    !WriteExact(rfbsock, (char *)&passwdLen, sizeof(passwdLen))) {
		return False;
	}

	if (!WriteExact(rfbsock, login, strlen(login)) ||
	    !WriteExact(rfbsock, passwd, strlen(passwd))) {
		return False;
	}

//	/* Lose the password from memory */
//	memset(passwd, '\0', strlen(passwd));

	if (!ReadFromRFBServer((char *)&authResult, sizeof(authResult))) {
		return False;
	}

	authResult = Swap32IfLE(authResult);

	switch (authResult) {
	case rfbVncAuthOK:
		fprintf(stderr, "Authentication succeeded\n\n");
		break;
	case rfbVncAuthFailed:
		fprintf(stderr, "Authentication failed\n\n");
		return False;
	case rfbVncAuthTooMany:
		fprintf(stderr, "Authentication failed - too many tries\n\n");
		return False;
	default:
		fprintf(stderr, "Unknown authentication result: %d\n\n",
		    (int)authResult);
		return False;
	}

	return True;
}


/*
 * In the protocol version 3.7t, the server informs us about supported
 * protocol messages and encodings. Here we read this information.
 */

static Bool
ReadInteractionCaps(void)
{
	rfbInteractionCapsMsg intr_caps;

	/* Read the counts of list items following */
	if (!ReadFromRFBServer((char *)&intr_caps, sz_rfbInteractionCapsMsg)) {
		return False;
	}
	intr_caps.nServerMessageTypes = Swap16IfLE(intr_caps.nServerMessageTypes);
	intr_caps.nClientMessageTypes = Swap16IfLE(intr_caps.nClientMessageTypes);
	intr_caps.nEncodingTypes = Swap16IfLE(intr_caps.nEncodingTypes);

	/* Read the lists of server- and client-initiated messages */
	return (ReadCapabilityList(serverMsgCaps, intr_caps.nServerMessageTypes) &&
		ReadCapabilityList(clientMsgCaps, intr_caps.nClientMessageTypes) &&
		ReadCapabilityList(encodingCaps, intr_caps.nEncodingTypes));
}


/*
 * Read the list of rfbCapabilityInfo structures and enable corresponding
 * capabilities in the specified container. The count argument specifies how
 * many records to read from the socket.
 */

static Bool
ReadCapabilityList(CapsContainer *caps, int count)
{
	rfbCapabilityInfo msginfo;
	int i;

	for (i = 0; i < count; i++) {
		if (!ReadFromRFBServer((char *)&msginfo, sz_rfbCapabilityInfo)) {
			return False;
		}
		msginfo.code = Swap32IfLE(msginfo.code);
		CapsEnable(caps, &msginfo);
	}

	return True;
}


/*
 * SetFormatAndEncodings.
 */

Bool
SetFormatAndEncodings()
{
  rfbSetPixelFormatMsg spf;
  char buf[sz_rfbSetEncodingsMsg + MAX_ENCODINGS * 4];
  rfbSetEncodingsMsg *se = (rfbSetEncodingsMsg *)buf;
  CARD32 *encs = (CARD32 *)(&buf[sz_rfbSetEncodingsMsg]);
  int len = 0;
  Bool requestCompressLevel = False;
  Bool requestQualityLevel = False;
  Bool requestLastRectEncoding = False;
  Bool requestNewFBSizeEncoding = True;
  Bool requestTextChatEncoding = True;

//  fprintf(stderr, "SetFormatAndEncodings: sent_FBU state: %2d\n", sent_FBU);

  spf.type = rfbSetPixelFormat;
  spf.format = myFormat;
  spf.format.redMax = Swap16IfLE(spf.format.redMax);
  spf.format.greenMax = Swap16IfLE(spf.format.greenMax);
  spf.format.blueMax = Swap16IfLE(spf.format.blueMax);


  if (!WriteExact(rfbsock, (char *)&spf, sz_rfbSetPixelFormatMsg))
    return False;

  se->type = rfbSetEncodings;
  se->nEncodings = 0;

  if (appData.encodingsString) {
    char *encStr = appData.encodingsString;
    int encStrLen;
    do {
      char *nextEncStr = strchr(encStr, ' ');
      if (nextEncStr) {
	encStrLen = nextEncStr - encStr;
	nextEncStr++;
      } else {
	encStrLen = strlen(encStr);
      }

if (getenv("DEBUG_SETFORMAT")) {
	fprintf(stderr, "encs: ");
	write(2, encStr, encStrLen);
	fprintf(stderr, "\n");
}

      if (strncasecmp(encStr,"raw",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);
      } else if (strncasecmp(encStr,"copyrect",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCopyRect);
      } else if (strncasecmp(encStr,"tight",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingTight);
	requestLastRectEncoding = True;
	if (appData.compressLevel >= 0 && appData.compressLevel <= 9)
	  requestCompressLevel = True;
	if (appData.enableJPEG)
	  requestQualityLevel = True;
      } else if (strncasecmp(encStr,"hextile",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingHextile);
      } else if (strncasecmp(encStr,"zlib",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZlib);
	if (appData.compressLevel >= 0 && appData.compressLevel <= 9)
	  requestCompressLevel = True;
      } else if (strncasecmp(encStr,"corre",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCoRRE);
      } else if (strncasecmp(encStr,"rre",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRRE);
      } else if (strncasecmp(encStr,"zrle",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZRLE);
      } else if (strncasecmp(encStr,"zywrle",encStrLen) == 0) {
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZYWRLE);
      } else {
	fprintf(stderr,"Unknown encoding '%.*s'\n",encStrLen,encStr);
      }

      encStr = nextEncStr;
    } while (encStr && se->nEncodings < MAX_ENCODINGS);

    if (se->nEncodings < MAX_ENCODINGS && requestCompressLevel) {
      encs[se->nEncodings++] = Swap32IfLE(appData.compressLevel +
					  rfbEncodingCompressLevel0);
    }

    if (se->nEncodings < MAX_ENCODINGS && requestQualityLevel) {
      if (appData.qualityLevel < 0 || appData.qualityLevel > 9)
        appData.qualityLevel = 5;
      encs[se->nEncodings++] = Swap32IfLE(appData.qualityLevel +
					  rfbEncodingQualityLevel0);
    }

    if (appData.useRemoteCursor) {
      if (se->nEncodings < MAX_ENCODINGS)
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingXCursor);
      if (se->nEncodings < MAX_ENCODINGS && !appData.useX11Cursor)
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRichCursor);
      if (se->nEncodings < MAX_ENCODINGS)
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingPointerPos);
    }

    if (se->nEncodings < MAX_ENCODINGS && requestLastRectEncoding) {
      encs[se->nEncodings++] = Swap32IfLE(rfbEncodingLastRect);
    }

    if (se->nEncodings < MAX_ENCODINGS && requestNewFBSizeEncoding) {
      encs[se->nEncodings++] = Swap32IfLE(rfbEncodingNewFBSize);
    }
  }
  else {
    if (SameMachine(rfbsock)) {
      if (!tunnelSpecified && appData.useRawLocal) {
	fprintf(stderr,"Same machine: preferring raw encoding\n");
	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);
      } else {
	fprintf(stderr,"Tunneling active: preferring tight encoding\n");
      }
    }

    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCopyRect);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingTight);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZRLE);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingHextile);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZlib);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCoRRE);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRRE);

    if (appData.compressLevel >= 0 && appData.compressLevel <= 9) {
      encs[se->nEncodings++] = Swap32IfLE(appData.compressLevel +
					  rfbEncodingCompressLevel0);
    } else if (!tunnelSpecified) {
      /* If -tunnel option was provided, we assume that server machine is
	 not in the local network so we use default compression level for
	 tight encoding instead of fast compression. Thus we are
	 requesting level 1 compression only if tunneling is not used. */
      encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCompressLevel1);
    }

    if (appData.enableJPEG) {
      if (appData.qualityLevel < 0 || appData.qualityLevel > 9)
	appData.qualityLevel = 5;
      encs[se->nEncodings++] = Swap32IfLE(appData.qualityLevel +
					  rfbEncodingQualityLevel0);
    }

    if (appData.useRemoteCursor) {
      encs[se->nEncodings++] = Swap32IfLE(rfbEncodingXCursor);
      if (!appData.useX11Cursor) {
      	encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRichCursor);
      }
      encs[se->nEncodings++] = Swap32IfLE(rfbEncodingPointerPos);
    }

    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingLastRect);
    encs[se->nEncodings++] = Swap32IfLE(rfbEncodingNewFBSize);
  }

  len = sz_rfbSetEncodingsMsg + se->nEncodings * 4;

  se->nEncodings = Swap16IfLE(se->nEncodings);

  if (!WriteExact(rfbsock, buf, len)) return False;

  return True;
}


/*
 * SendIncrementalFramebufferUpdateRequest.
 */

Bool
SendIncrementalFramebufferUpdateRequest()
{
	return SendFramebufferUpdateRequest(0, 0, si.framebufferWidth,
	    si.framebufferHeight, True);
}

time_t last_filexfer = 0;

/*
 * SendFramebufferUpdateRequest.
 */

Bool
SendFramebufferUpdateRequest(int x, int y, int w, int h, Bool incremental)
{
	rfbFramebufferUpdateRequestMsg fur;

	if (appData.fileActive) {
		if (time(NULL) < last_filexfer + 2) {
//			fprintf(stderr, "skip SendFramebufferUpdateRequest: %d - %d\n", last_filexfer, time(NULL));
//			return True;
			x = y = 0;
			w = h = 1;
		}
	}

	fur.type = rfbFramebufferUpdateRequest;
	fur.incremental = incremental ? 1 : 0;
	fur.x = Swap16IfLE(x);
	fur.y = Swap16IfLE(y);
	fur.w = Swap16IfLE(w);
	fur.h = Swap16IfLE(h);

	if (incremental) {
		sent_FBU = 1;
	} else {
		sent_FBU = 2;
	}

	if (!WriteExact(rfbsock, (char *)&fur, sz_rfbFramebufferUpdateRequestMsg)) {
		return False;
	}

	return True;
}


/*
 * SendPointerEvent.
 */

Bool
SendPointerEvent(int x, int y, int buttonMask)
{
	rfbPointerEventMsg pe;

	if (appData.fileActive) {
		if (time(NULL) < last_filexfer + 2) {
			//fprintf(stderr, "skip SendPointerEvent: %d - %d\n", last_filexfer, time(NULL));
			return True;
		}
	}

	pe.type = rfbPointerEvent;
	pe.buttonMask = buttonMask;
	if (x < 0) x = 0;
	if (y < 0) y = 0;

	if (!appData.useX11Cursor) {
		SoftCursorMove(x, y);
	}

	pe.x = Swap16IfLE(x);
	pe.y = Swap16IfLE(y);
	return WriteExact(rfbsock, (char *)&pe, sz_rfbPointerEventMsg);
}


/*
 * SendKeyEvent.
 */

Bool
SendKeyEvent(CARD32 key, Bool down)
{
	rfbKeyEventMsg ke;

	ke.type = rfbKeyEvent;
	ke.down = down ? 1 : 0;
	ke.key = Swap32IfLE(key);
	return WriteExact(rfbsock, (char *)&ke, sz_rfbKeyEventMsg);
}


/*
 * SendClientCutText.
 */

Bool
SendClientCutText(char *str, int len)
{
	rfbClientCutTextMsg cct;

	if (serverCutText) {
		free(serverCutText);
	}
	serverCutText = NULL;

	if (appData.viewOnly) {
		return True;
	}

	cct.type = rfbClientCutText;
	cct.length = Swap32IfLE(len);
	return  (WriteExact(rfbsock, (char *)&cct, sz_rfbClientCutTextMsg) &&
	    WriteExact(rfbsock, str, len));
}

Bool
SendServerScale(int nfac)
{
	rfbSetScaleMsg ssc;
	if (nfac < 0 || nfac > 100) {
		return True;
	}

	ssc.type = rfbSetScale;
	ssc.scale = nfac;
	return WriteExact(rfbsock, (char *)&ssc, sz_rfbSetScaleMsg);
}

Bool
SendServerInput(Bool enabled)
{
	rfbSetServerInputMsg sim;

	sim.type = rfbSetServerInput;
	sim.status = enabled;
	return WriteExact(rfbsock, (char *)&sim, sz_rfbSetServerInputMsg);
}

Bool
SendSingleWindow(int x, int y)
{
	rfbSetSWMsg sw;

	fprintf(stderr, "SendSingleWindow: %d %d\n", x, y);

	if (x == -1 && y == -1)  {
		sw.type = rfbSetSW;
		sw.x = Swap16IfLE(1);
		sw.y = Swap16IfLE(1);
	} else {
		sw.type = rfbSetSW;
		sw.x = Swap16IfLE(x);
		sw.y = Swap16IfLE(y);
	}
	sw.status = True;
	return WriteExact(rfbsock, (char *)&sw, sz_rfbSetSWMsg);
}

Bool
SendTextChat(char *str)
{
	rfbTextChatMsg chat;
	chat.type = rfbTextChat;
	chat.pad1 = 0;
	chat.pad2 = 0;
	chat.length = (unsigned int) strlen(str);
	//fprintf(stderr, "SendTextChat: %d '%s'\n", chat.length, str);
	chat.length = Swap32IfLE(chat.length);
	if (!WriteExact(rfbsock, (char *)&chat, sz_rfbTextChatMsg)) {
		return False;
	}
	return WriteExact(rfbsock, str, strlen(str));
}

extern void raiseme(void);

Bool
SendTextChatOpen(void)
{
	raiseme();
	rfbTextChatMsg chat;
	chat.type = rfbTextChat;
	chat.pad1 = 0;
	chat.pad2 = 0;
	chat.length = Swap32IfLE(rfbTextChatOpen);
	return WriteExact(rfbsock, (char *)&chat, sz_rfbTextChatMsg);
}

Bool
SendTextChatClose(void)
{
	rfbTextChatMsg chat;
	chat.type = rfbTextChat;
	chat.pad1 = 0;
	chat.pad2 = 0;
	chat.length = Swap32IfLE(rfbTextChatClose);
	return WriteExact(rfbsock, (char *)&chat, sz_rfbTextChatMsg);
}

Bool
SendTextChatFinished(void)
{
	rfbTextChatMsg chat;
	chat.type = rfbTextChat;
	chat.pad1 = 0;
	chat.pad2 = 0;
	chat.length = Swap32IfLE(rfbTextChatFinished);
	return WriteExact(rfbsock, (char *)&chat, sz_rfbTextChatMsg);
}

extern int do_format_change;
extern int do_cursor_change;
extern double do_fb_update;
extern void cutover_format_change(void);

double dtime(double *t_old) {
        /* 
         * usage: call with 0.0 to initialize, subsequent calls give
         * the time difference since last call.
         */
        double t_now, dt;
        struct timeval now;

        gettimeofday(&now, NULL);
        t_now = now.tv_sec + ( (double) now.tv_usec/1000000. );
        if (*t_old == 0.0) {
                *t_old = t_now;
                return t_now;
        }
        dt = t_now - *t_old;
        *t_old = t_now;
        return(dt);
}

/* common dtime() activities: */
double dtime0(double *t_old) {
        *t_old = 0.0;
        return dtime(t_old);
}

double dnow(void) {
        double t;
        return dtime0(&t);
}

extern int filexfer_sock;
extern int filexfer_listen;

static char fxfer[65536];
extern void CheckFileXfer(void);

Bool HandleFileXfer(void) {
	unsigned char hdr[12];
	unsigned int len;

        int rfbDirContentRequest = 1;
        int rfbDirPacket = 2; // Full directory name or full file name.
        int rfbFileTransferRequest = 3;
        int rfbFileHeader = 4;
        int rfbFilePacket = 5; // One slice of the file
        int rfbEndOfFile = 6;
        int rfbAbortFileTransfer = 7;
        int rfbFileTransferOffer = 8;
        int rfbFileAcceptHeader = 9; // The server accepts or rejects the file
        int rfbCommand = 10;
        int rfbCommandReturn = 11;
        int rfbFileChecksums = 12;

        int rfbRDirContent = 1; // Request a Server Directory contents
        int rfbRDrivesList = 2; // Request the server's drives list

        int rfbADirectory = 1; // Reception of a directory name
        int rfbAFile = 2; // Reception of a file name 
        int rfbADrivesList = 3; // Reception of a list of drives
        int rfbADirCreate = 4; // Response to a create dir command 
        int rfbADirDelete = 5; // Response to a delete dir command 
        int rfbAFileCreate = 6; // Response to a create file command 
        int rfbAFileDelete = 7; // Response to a delete file command

        int rfbCDirCreate = 1; // Request the server to create the given directory
        int rfbCDirDelete = 2; // Request the server to delete the given directory
        int rfbCFileCreate = 3; // Request the server to create the given file
        int rfbCFileDelete = 4; // Request the server to delete the given file

        int rfbRErrorUnknownCmd = 1; // Unknown FileTransfer command.
        int rfbRErrorCmd = 0xFFFFFFFF;

	int db = 0;

	last_filexfer = time(NULL);
	//fprintf(stderr, "last_filexfer-1: %d\n", last_filexfer);

	if (filexfer_sock < 0) {
		return True;
	}

	hdr[0] = rfbFileTransfer;
	skip_XtUpdateAll = 1;
	if (!ReadFromRFBServer(&hdr[1], 11)) {
		skip_XtUpdateAll = 0;
		return False;
	}
	write(filexfer_sock, hdr, 12);
	if (db) fprintf(stderr, "Got rfbFileTransfer hdr\n");
	if (db) write(2, hdr, 12);
	if (db) fprintf(stderr, "\n");

	len = (hdr[8] << 24) | (hdr[9] << 16) | (hdr[10] << 8) | hdr[11];
	if (db) fprintf(stderr, "Got rfbFileTransfer: len %d\n", len);
	if (len > 0) {
		if (!ReadFromRFBServer(fxfer, len)) {
			skip_XtUpdateAll = 0;
			return False;
		}
		if (db) write(2, fxfer, len);
		if (db) fprintf(stderr, "\n");
		write(filexfer_sock, fxfer, len);
	}
	len = (hdr[4] << 24) | (hdr[5] << 16) | (hdr[6] << 8) | hdr[7];
	if (hdr[1] == rfbFileHeader && len != rfbRErrorCmd) {
		if (db) fprintf(stderr, "Got rfbFileTransfer: rfbFileHeader\n");
		len = 4;
		if (!ReadFromRFBServer(fxfer, len)) {
			skip_XtUpdateAll = 0;
			return False;
		}
		if (db) write(2, fxfer, len);
		if (db) fprintf(stderr, "\n");
		write(filexfer_sock, fxfer, len);
	}
	if (db) fprintf(stderr, "Got rfbFileTransfer done.\n");
	skip_XtUpdateAll = 0;

	CheckFileXfer();
	return True;
}

/*
 * HandleRFBServerMessage.
 */


Bool
HandleRFBServerMessage()
{
	int db = 0;
	rfbServerToClientMsg msg;

	if (!ReadFromRFBServer((char *)&msg, 1)) {
		return False;
	}

	if (msg.type == rfbFileTransfer) {
		return HandleFileXfer();
	}

    switch (msg.type) {

    case rfbSetColourMapEntries:
    {
	int i;
	CARD16 rgb[3];
	XColor xc;

	if (!ReadFromRFBServer(((char *)&msg) + 1, sz_rfbSetColourMapEntriesMsg - 1)) {
		return False;
	}

	msg.scme.firstColour = Swap16IfLE(msg.scme.firstColour);
	msg.scme.nColours = Swap16IfLE(msg.scme.nColours);

	for (i = 0; i < msg.scme.nColours; i++) {
		if (!ReadFromRFBServer((char *)rgb, 6)) {
			return False;
		}
		xc.pixel = msg.scme.firstColour + i;
		xc.red = Swap16IfLE(rgb[0]);
		xc.green = Swap16IfLE(rgb[1]);
		xc.blue = Swap16IfLE(rgb[2]);
		if (appData.useGreyScale) {
			int ave = (xc.red + xc.green + xc.blue)/3;
			xc.red   = ave;
			xc.green = ave;
			xc.blue  = ave;
		}
		xc.flags = DoRed|DoGreen|DoBlue;
		XStoreColor(dpy, cmap, &xc);
	}

	break;
    }

    case rfbFramebufferUpdate:
    {
	rfbFramebufferUpdateRectHeader rect;
	int linesToRead;
	int bytesPerLine;
	int i;
	int usecs;

	int area_copyrect = 0;
	int area_tight = 0;
	int area_zrle = 0;
	int area_raw = 0;
	if (db) fprintf(stderr, "FBU-0:    %.6f\n", dnow());

	int skip_incFBU = 0;
	sent_FBU = -1;

	if (!ReadFromRFBServer(((char *)&msg.fu) + 1, sz_rfbFramebufferUpdateMsg - 1)) {
		return False;
	}

	msg.fu.nRects = Swap16IfLE(msg.fu.nRects);

	for (i = 0; i < msg.fu.nRects; i++) {
		if (!ReadFromRFBServer((char *)&rect, sz_rfbFramebufferUpdateRectHeader)) {
			return False;
		}

		rect.encoding = Swap32IfLE(rect.encoding);
		if (rect.encoding == rfbEncodingLastRect) {
			break;
		}

		rect.r.x = Swap16IfLE(rect.r.x);
		rect.r.y = Swap16IfLE(rect.r.y);
		rect.r.w = Swap16IfLE(rect.r.w);
		rect.r.h = Swap16IfLE(rect.r.h);

		if (rect.encoding == rfbEncodingXCursor || rect.encoding == rfbEncodingRichCursor) {
			if (db) fprintf(stderr, "FBU-Cur1  %.6f\n", dnow());
			if (!HandleCursorShape(rect.r.x, rect.r.y, rect.r.w, rect.r.h, rect.encoding)) {
				return False;
			}
			if (db) fprintf(stderr, "FBU-Cur2  %.6f\n", dnow());
			continue;
		}

		if (rect.encoding == rfbEncodingPointerPos) {
			if (db) fprintf(stderr, "FBU-Pos1  %.6f\n", dnow());
			if (!HandleCursorPos(rect.r.x, rect.r.y)) {
				return False;
			}
			if (db) fprintf(stderr, "FBU-Pos2  %.6f\n", dnow());
			continue;
		}
		if (rect.encoding == rfbEncodingNewFBSize) {
			fprintf(stderr,"New Size: %dx%d at (%d, %d)\n", rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			si.framebufferWidth = rect.r.w;
			si.framebufferHeight = rect.r.h;
			//fprintf(stderr, "si: %d %d\n", si.framebufferWidth, si.framebufferHeight);
			ReDoDesktop();
			continue;
		}

		if ((rect.r.x + rect.r.w > si.framebufferWidth) ||
		    (rect.r.y + rect.r.h > si.framebufferHeight)) {
			fprintf(stderr,"Rect too large: %dx%d at (%d, %d)\n",
			rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			return False;
		}

		if (rect.r.h * rect.r.w == 0) {
			fprintf(stderr,"Zero size rect - ignoring\n");
			continue;
		}

		/* If RichCursor encoding is used, we should prevent collisions
		   between framebuffer updates and cursor drawing operations. */
		if (db) fprintf(stderr, "FBU-SCL1  %.6f\n", dnow());

		SoftCursorLockArea(rect.r.x, rect.r.y, rect.r.w, rect.r.h);

		if (db) fprintf(stderr, "FBU-SCL2  %.6f\n", dnow());

		switch (rect.encoding) {

		case rfbEncodingRaw:

			bytesPerLine = rect.r.w * myFormat.bitsPerPixel / 8;
			linesToRead = BUFFER_SIZE / bytesPerLine;

			if (db) fprintf(stderr, "Raw:     %dx%d+%d+%d\n", rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			area_raw += rect.r.w * rect.r.h;

			while (rect.r.h > 0) {
				if (linesToRead > rect.r.h) {
					linesToRead = rect.r.h;
				}

				if (!ReadFromRFBServer(buffer,bytesPerLine * linesToRead)) {
					return False;
				}

				CopyDataToScreen(buffer, rect.r.x, rect.r.y, rect.r.w, linesToRead);

				rect.r.h -= linesToRead;
				rect.r.y += linesToRead;
			}
			break;

		case rfbEncodingCopyRect:
		{
			rfbCopyRect cr;

			if (!ReadFromRFBServer((char *)&cr, sz_rfbCopyRect)) {
				return False;
			}

			cr.srcX = Swap16IfLE(cr.srcX);
			cr.srcY = Swap16IfLE(cr.srcY);

			if (db) fprintf(stderr, "Copy:    %dx%d+%d+%d\n", rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			area_copyrect += rect.r.w * rect.r.h;

			/* If RichCursor encoding is used, we should extend our
			   "cursor lock area" (previously set to destination
			   rectangle) to the source rectangle as well. */

			if (db) fprintf(stderr, "FBU-SCL3  %.6f\n", dnow());

			SoftCursorLockArea(cr.srcX, cr.srcY, rect.r.w, rect.r.h);

			if (db) fprintf(stderr, "FBU-SCL4  %.6f\n", dnow());

			if (appData.copyRectDelay != 0) {
				XFillRectangle(dpy, desktopWin, srcGC, cr.srcX, cr.srcY, rect.r.w, rect.r.h);
				XFillRectangle(dpy, desktopWin, dstGC, rect.r.x, rect.r.y, rect.r.w, rect.r.h);
				XSync(dpy,False);
				usleep(appData.copyRectDelay * 1000);
				XFillRectangle(dpy, desktopWin, dstGC, rect.r.x, rect.r.y, rect.r.w, rect.r.h);
				XFillRectangle(dpy, desktopWin, srcGC, cr.srcX, cr.srcY, rect.r.w, rect.r.h);
			}

			if (db) fprintf(stderr, "FBU-CPA1  %.6f\n", dnow());
			if (!appData.useBackingstore) {
				copy_rect(rect.r.x, rect.r.y, rect.r.w, rect.r.h, cr.srcX, cr.srcY);
				put_image(rect.r.x, rect.r.y, rect.r.x, rect.r.y, rect.r.w, rect.r.h);
				XSync(dpy, False);
			} else {
				XCopyArea(dpy, desktopWin, desktopWin, gc, cr.srcX, cr.srcY,
				    rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			}
			if (db) fprintf(stderr, "FBU-CPA2  %.6f\n", dnow());

			break;
		}

		case rfbEncodingRRE:
		{
			switch (myFormat.bitsPerPixel) {
			case 8:
				if (!HandleRRE8(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 16:
				if (!HandleRRE16(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 32:
				if (!HandleRRE32(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			}
			break;
		}

		case rfbEncodingCoRRE:
		{
			switch (myFormat.bitsPerPixel) {
			case 8:
				if (!HandleCoRRE8(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 16:
				if (!HandleCoRRE16(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 32:
				if (!HandleCoRRE32(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			}
			break;
		}

		case rfbEncodingHextile:
		{
			switch (myFormat.bitsPerPixel) {
			case 8:
				if (!HandleHextile8(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 16:
				if (!HandleHextile16(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 32:
				if (!HandleHextile32(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			}
			break;
		}

		case rfbEncodingZlib:
		{
			switch (myFormat.bitsPerPixel) {
			case 8:
				if (!HandleZlib8(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 16:
				if (!HandleZlib16(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 32:
				if (!HandleZlib32(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			}
			break;
		}

		case rfbEncodingTight:
		{
			if (db) fprintf(stderr, "Tight:   %dx%d+%d+%d\n", rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			area_tight += rect.r.w * rect.r.h;
			if (db) fprintf(stderr, "FBU-TGH1  %.6f\n", dnow());

			switch (myFormat.bitsPerPixel) {
			case 8:
				if (!HandleTight8(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 16:
				if (!HandleTight16(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 32:
				if (!HandleTight32(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			}
			if (db) fprintf(stderr, "FBU-TGH2  %.6f\n", dnow());
			break;
		}

		/* runge adds zrle and zywrle: */
		case rfbEncodingZRLE:
		{
			if (db) fprintf(stderr, "ZRLE:    %dx%d+%d+%d\n", rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			area_zrle += rect.r.w * rect.r.h;
			switch (myFormat.bitsPerPixel) {
			case 8:
				if (!HandleZRLE8(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 16:
				if (!HandleZRLE16(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			case 32:
			    {
				unsigned int maxColor=(myFormat.redMax<<myFormat.redShift)|
				    (myFormat.greenMax<<myFormat.greenShift)|
				    (myFormat.blueMax<<myFormat.blueShift);
				if ((myFormat.bigEndian && (maxColor&0xff)==0) || (!myFormat.bigEndian && (maxColor&0xff000000)==0)) {
					if (!HandleZRLE24(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
						return False;
					}
				} else if (!myFormat.bigEndian && (maxColor&0xff)==0) {
					if (!HandleZRLE24Up(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
						return False;
					}
				} else if (myFormat.bigEndian && (maxColor&0xff000000)==0) {
					if (!HandleZRLE24Down(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
						return False;
					}
				} else if (!HandleZRLE32(rect.r.x,rect.r.y,rect.r.w,rect.r.h)) {
					return False;
				}
				break;
			    }
			}
			break;
		}

		default:
		fprintf(stderr,"Unknown rect encoding %d\n", (int)rect.encoding);
		return False;
		}

		/* Now we may discard "soft cursor locks". */
		if (db) fprintf(stderr, "FBU-SUL1  %.6f\n", dnow());

		SoftCursorUnlockScreen();

		if (db) fprintf(stderr, "FBU-SUL2  %.6f\n", dnow());
	}

	if (1 || area_copyrect) {
		if (db) fprintf(stderr, "FBU-XSN1  %.6f\n", dnow());
		XSync(dpy, False);
		if (db) fprintf(stderr, "FBU-XSN2  %.6f\n", dnow());
	}
	sent_FBU = 0;
	/*
	 * we need to be careful since Xt events are processed
	 * usually in the middle of FBU.  So we do any scheduled ones now
	 * which is pretty safe but not absolutely safe.
         */
	if (do_format_change) {
		cutover_format_change();
		do_format_change = 0;
		SetVisualAndCmap();
		SetFormatAndEncodings();
		if (do_cursor_change) {
			if (do_cursor_change == 1) {
				DesktopCursorOff();
			}
			do_cursor_change = 0;
		} else {
			SendFramebufferUpdateRequest(0, 0, si.framebufferWidth,
			    si.framebufferHeight, False);
			skip_incFBU = 1;
		}
	}
	if (do_fb_update != 0.0) {
		if (dnow() > do_fb_update + 1.1) {
			do_fb_update = 0.0;
			SendFramebufferUpdateRequest(0, 0, si.framebufferWidth,
			    si.framebufferHeight, False);
		}
	}

#ifdef MITSHM
    /* if using shared memory PutImage, make sure that the X server has
       updated its framebuffer before we reuse the shared memory.  This is
       mainly to avoid copyrect using invalid screen contents - not sure
       if we'd need it otherwise. */

	if (appData.useShm) {
		XSync(dpy, False);
	} else
#endif
	{
		XSync(dpy, False);
	}
	
	if (skip_XtUpdate || skip_incFBU) {
		;
	} else if (!SendIncrementalFramebufferUpdateRequest()) {
		return False;
	}

	break;
  }

  case rfbBell:
  {
	Window toplevelWin;

	if (appData.useBell) {
		XBell(dpy, 0);
	}

	if (appData.raiseOnBeep) {
		toplevelWin = XtWindow(toplevel);
		XMapRaised(dpy, toplevelWin);
	}

	break;
    }

    case rfbServerCutText:
    {
	if (!ReadFromRFBServer(((char *)&msg) + 1, sz_rfbServerCutTextMsg - 1)) {
		return False;
	}

	msg.sct.length = Swap32IfLE(msg.sct.length);

	if (serverCutText) {
		free(serverCutText);
	}

	serverCutText = malloc(msg.sct.length+1);

	if (!ReadFromRFBServer(serverCutText, msg.sct.length)) {
		return False;
	}

	serverCutText[msg.sct.length] = 0;

	newServerCutText = True;

	break;
    }

    case rfbTextChat:
    {
	char *buffer = NULL;
	if (!ReadFromRFBServer(((char *)&msg) + 1, sz_rfbTextChatMsg - 1)) {
		return False;
	}
	msg.tc.length = Swap32IfLE(msg.tc.length);
	switch(msg.tc.length) {
	case rfbTextChatOpen:
		if (appData.termChat) {
			printChat("\n*ChatOpen*\n\nSend: ", True);
		} else {
			printChat("\n*ChatOpen*\n", True);
		}
		appData.chatActive = True;
		break;
	case rfbTextChatClose:
		printChat("\n*ChatClose*\n", False);
		appData.chatActive = False;
		break;
	case rfbTextChatFinished:
		printChat("\n*ChatFinished*\n", False);
		appData.chatActive = False;
		break;
	default:
		buffer = (char *)malloc(msg.tc.length+1);
		if (!ReadFromRFBServer(buffer, msg.tc.length)) {
			free(buffer);
			return False;
		}
		buffer[msg.tc.length] = '\0';
		appData.chatActive = True;
		GotChatText(buffer, msg.tc.length);
		free(buffer);
	}
	break;
    }

    case rfbRestartConnection:
    {
	rfbRestartConnectionMsg rc;
	int len;
	char *rs_str;
	char buf[5] = "\xff\xff\xff\xff";
	fprintf(stderr, "rfbRestartConnection. type=%d\n", (int) rc.type);
	if (!ReadFromRFBServer((char *)&rc + 1, sz_rfbRestartConnectionMsg - 1)) {
		return False;
	}
	len = Swap32IfLE(rc.length);
	fprintf(stderr, "rfbRestartConnection. pad1=%d\n", (int) rc.pad1);
	fprintf(stderr, "rfbRestartConnection. pad2=%d\n", (int) rc.pad2);
	fprintf(stderr, "rfbRestartConnection. len=%d\n", len);
	if (len) {
		rs_str = (char *)malloc(2*len);	
		if (!ReadFromRFBServer(rs_str, len)) {
			return False;
		}
		restart_session_pw = rs_str;
		restart_session_len = len;
	}
	if (!WriteExact(rfbsock, buf, 4)) {
		return False;
	}
	InitialiseRFBConnection();
	SetVisualAndCmap();
	SetFormatAndEncodings();
	DesktopCursorOff();
	SendFramebufferUpdateRequest(0, 0, si.framebufferWidth, si.framebufferHeight, False);

	break;
    }

    default:
	fprintf(stderr,"Unknown message type %d from VNC server\n",msg.type);
	return False;
    }
	if (appData.fileActive) {
		if (filexfer_sock < 0 && filexfer_listen < 0) {
			appData.fileActive = False;
		} else {
			CheckFileXfer();
		}
	}

    return True;
}


#define GET_PIXEL8(pix, ptr) ((pix) = *(ptr)++)

#define GET_PIXEL16(pix, ptr) (((CARD8*)&(pix))[0] = *(ptr)++, \
			       ((CARD8*)&(pix))[1] = *(ptr)++)

#define GET_PIXEL32(pix, ptr) (((CARD8*)&(pix))[0] = *(ptr)++, \
			       ((CARD8*)&(pix))[1] = *(ptr)++, \
			       ((CARD8*)&(pix))[2] = *(ptr)++, \
			       ((CARD8*)&(pix))[3] = *(ptr)++)

/* CONCAT2 concatenates its two arguments.  CONCAT2E does the same but also
   expands its arguments if they are macros */

#define CONCAT2(a,b) a##b
#define CONCAT2E(a,b) CONCAT2(a,b)

#define CONCAT3(a,b,c) a##b##c
#define CONCAT3E(a,b,c) CONCAT3(a,b,c)

static unsigned char* frameBuffer = NULL;
static int frameBufferLen = 0;

#define BPP 8
#include "rre.c"
#include "corre.c"
#include "hextile.c"
#include "zlib.c"
#include "tight.c"
#include "zrle.c"
#undef BPP
#define BPP 16
#include "rre.c"
#include "corre.c"
#include "hextile.c"
#include "zlib.c"
#include "tight.c"
#include "zrle.c"
#undef BPP
#define BPP 32
#include "rre.c"
#include "corre.c"
#include "hextile.c"
#include "zlib.c"
#include "tight.c"
#include "zrle.c"
#define REALBPP 24
#include "zrle.c"
#define REALBPP 24
#define UNCOMP 8
#include "zrle.c"
#define REALBPP 24
#define UNCOMP -8
#include "zrle.c"
#undef BPP

/*
 * Read the string describing the reason for a connection failure.
 */

static void
ReadConnFailedReason(void)
{
  CARD32 reasonLen;
  char *reason = NULL;

  if (ReadFromRFBServer((char *)&reasonLen, sizeof(reasonLen))) {
    reasonLen = Swap32IfLE(reasonLen);
    if ((reason = malloc(reasonLen)) != NULL &&
        ReadFromRFBServer(reason, reasonLen)) {
      fprintf(stderr,"VNC connection failed: %.*s\n", (int)reasonLen, reason);
      free(reason);
      return;
    }
  }

  fprintf(stderr, "VNC connection failed\n");

  if (reason != NULL)
    free(reason);
}

/*
 * PrintPixelFormat.
 */

void
PrintPixelFormat(format)
    rfbPixelFormat *format;
{
  if (format->bitsPerPixel == 1) {
    fprintf(stderr,"  Single bit per pixel.\n");
    fprintf(stderr,
	    "  %s significant bit in each byte is leftmost on the screen.\n",
	    (format->bigEndian ? "Most" : "Least"));
  } else {
    fprintf(stderr,"  %d bits per pixel.  ",format->bitsPerPixel);
    if (format->bitsPerPixel != 8) {
      fprintf(stderr,"%s significant byte first in each pixel.\n",
	      (format->bigEndian ? "Most" : "Least"));
    }
    if (format->trueColour) {
      fprintf(stderr,"  True colour: max red %d green %d blue %d",
	      format->redMax, format->greenMax, format->blueMax);
      fprintf(stderr,", shift red %d green %d blue %d\n",
	      format->redShift, format->greenShift, format->blueShift);
    } else {
      fprintf(stderr,"  Colour map (not true colour).\n");
    }
  }
}

/*
 * Read an integer value encoded in 1..3 bytes. This function is used
 * by the Tight decoder.
 */

static long
ReadCompactLen (void)
{
  long len;
  CARD8 b;

  if (!ReadFromRFBServer((char *)&b, 1))
    return -1;
  len = (int)b & 0x7F;
  if (b & 0x80) {
    if (!ReadFromRFBServer((char *)&b, 1))
      return -1;
    len |= ((int)b & 0x7F) << 7;
    if (b & 0x80) {
      if (!ReadFromRFBServer((char *)&b, 1))
	return -1;
      len |= ((int)b & 0xFF) << 14;
    }
  }
  return len;
}


/*
 * JPEG source manager functions for JPEG decompression in Tight decoder.
 */

static struct jpeg_source_mgr jpegSrcManager;
static JOCTET *jpegBufferPtr;
static size_t jpegBufferLen;

static void
JpegInitSource(j_decompress_ptr cinfo)
{
  jpegError = False;
}

static boolean
JpegFillInputBuffer(j_decompress_ptr cinfo)
{
  jpegError = True;
  jpegSrcManager.bytes_in_buffer = jpegBufferLen;
  jpegSrcManager.next_input_byte = (JOCTET *)jpegBufferPtr;

  return TRUE;
}

static void
JpegSkipInputData(j_decompress_ptr cinfo, long num_bytes)
{
  if (num_bytes < 0 || num_bytes > jpegSrcManager.bytes_in_buffer) {
    jpegError = True;
    jpegSrcManager.bytes_in_buffer = jpegBufferLen;
    jpegSrcManager.next_input_byte = (JOCTET *)jpegBufferPtr;
  } else {
    jpegSrcManager.next_input_byte += (size_t) num_bytes;
    jpegSrcManager.bytes_in_buffer -= (size_t) num_bytes;
  }
}

static void
JpegTermSource(j_decompress_ptr cinfo)
{
  /* No work necessary here. */
}

static void
JpegSetSrcManager(j_decompress_ptr cinfo, CARD8 *compressedData,
		  int compressedLen)
{
  jpegBufferPtr = (JOCTET *)compressedData;
  jpegBufferLen = (size_t)compressedLen;

  jpegSrcManager.init_source = JpegInitSource;
  jpegSrcManager.fill_input_buffer = JpegFillInputBuffer;
  jpegSrcManager.skip_input_data = JpegSkipInputData;
  jpegSrcManager.resync_to_restart = jpeg_resync_to_restart;
  jpegSrcManager.term_source = JpegTermSource;
  jpegSrcManager.next_input_byte = jpegBufferPtr;
  jpegSrcManager.bytes_in_buffer = jpegBufferLen;

  cinfo->src = &jpegSrcManager;
}

#define DO_ZYWRLE 0
#if DO_ZYWRLE

#define ZYWRLE_ENDIAN ENDIAN_LITTLE

#undef END_FIX
#if ZYWRLE_ENDIAN == ENDIAN_LITTLE
#  define END_FIX LE
#elif ZYWRLE_ENDIAN == ENDIAN_BIG
#  define END_FIX BE
#else
#  define END_FIX NE
#endif
#define ZYWRLE_DECODE

#ifndef __RFB_CONCAT2E
#define __RFB_CONCAT2(a,b) a##b
#define __RFB_CONCAT2E(a,b) __RFB_CONCAT2(a,b)
#endif

#ifndef __RFB_CONCAT3E
#define __RFB_CONCAT3(a,b,c) a##b##c
#define __RFB_CONCAT3E(a,b,c) __RFB_CONCAT3(a,b,c)
#endif

#undef END_FIX
#if ZYWRLE_ENDIAN == ENDIAN_LITTLE
#  define END_FIX LE
#elif ZYWRLE_ENDIAN == ENDIAN_BIG
#  define END_FIX BE
#else
#  define END_FIX NE
#endif

#include "zrletypes.h"
#include "zrlepalettehelper.h"

#if 0
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#ifdef CPIXEL
#define PIXEL_T __RFB_CONCAT3E(uint,BPP,_t)
#define zrleOutStreamWRITE_PIXEL __RFB_CONCAT2E(zrleOutStreamWriteOpaque,CPIXEL)
#define ZRLE_ENCODE __RFB_CONCAT3E(zrleEncode,CPIXEL,END_FIX)
#define ZRLE_ENCODE_TILE __RFB_CONCAT3E(zrleEncodeTile,CPIXEL,END_FIX)
#define BPPOUT 24
#elif BPP==15
#define PIXEL_T __RFB_CONCAT3E(uint,16,_t)
#define zrleOutStreamWRITE_PIXEL __RFB_CONCAT2E(zrleOutStreamWriteOpaque,16)
#define ZRLE_ENCODE __RFB_CONCAT3E(zrleEncode,BPP,END_FIX)
#define ZRLE_ENCODE_TILE __RFB_CONCAT3E(zrleEncodeTile,BPP,END_FIX)
#define BPPOUT 16
#else
#define PIXEL_T __RFB_CONCAT3E(uint,BPP,_t)
#define zrleOutStreamWRITE_PIXEL __RFB_CONCAT2E(zrleOutStreamWriteOpaque,BPP)
#define ZRLE_ENCODE __RFB_CONCAT3E(zrleEncode,BPP,END_FIX)
#define ZRLE_ENCODE_TILE __RFB_CONCAT3E(zrleEncodeTile,BPP,END_FIX)
#define BPPOUT BPP
#endif

#else

#ifdef CPIXEL
#define PIXEL_T __RFB_CONCAT2E(zrle_U,BPP)
#define zrleOutStreamWRITE_PIXEL __RFB_CONCAT2E(zrleOutStreamWriteOpaque,CPIXEL)
#define ZRLE_ENCODE __RFB_CONCAT3E(zrleEncode,CPIXEL,END_FIX)
#define ZRLE_ENCODE_TILE __RFB_CONCAT3E(zrleEncodeTile,CPIXEL,END_FIX)
#define BPPOUT 24
#elif BPP==15
#define PIXEL_T __RFB_CONCAT2E(zrle_U,16)
#define zrleOutStreamWRITE_PIXEL __RFB_CONCAT2E(zrleOutStreamWriteOpaque,16)
#define ZRLE_ENCODE __RFB_CONCAT3E(zrleEncode,BPP,END_FIX)
#define ZRLE_ENCODE_TILE __RFB_CONCAT3E(zrleEncodeTile,BPP,END_FIX)
#define BPPOUT 16
#else
#define PIXEL_T __RFB_CONCAT2E(zrle_U,BPP)
#define zrleOutStreamWRITE_PIXEL __RFB_CONCAT2E(zrleOutStreamWriteOpaque,BPP)
#define ZRLE_ENCODE __RFB_CONCAT3E(zrleEncode,BPP,END_FIX)
#define ZRLE_ENCODE_TILE __RFB_CONCAT3E(zrleEncodeTile,BPP,END_FIX)
#define BPPOUT BPP
#endif

#endif

#ifndef ZRLE_ONCE
#define ZRLE_ONCE

static const int bitsPerPackedPixel[] = {
  0, 1, 2, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4
};

int zywrle_level;
int zywrleBuf[rfbZRLETileWidth*rfbZRLETileHeight];

static zrlePaletteHelper paletteHelper;

#endif /* ZRLE_ONCE */

#if 0
void ZRLE_ENCODE_TILE (PIXEL_T* data, int w, int h, zrleOutStream* os);
#endif

#define BPP 16
#include "zywrletemplate.c"
#undef BPP
#define BPP 32
#include "zywrletemplate.c"
#undef BPP

#endif	/* DO_ZYWRLE */

