/*
 *  Copyright (C) 2005 Johannes E. Schindelin.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * zrle.c - handle zrle encoding.
 *
 * This file shouldn't be compiled directly.  It is included multiple times by
 * rfbproto.c, each time with a different definition of the macro BPP.  For
 * each value of BPP, this file defines a function which handles an zrle
 * encoded rectangle with BPP bits per pixel.
 */

#ifndef REALBPP
#define REALBPP BPP
#endif

#if !defined(UNCOMP) || UNCOMP==0
#define HandleZRLE CONCAT2E(HandleZRLE,REALBPP)
#define HandleZRLETile CONCAT2E(HandleZRLETile,REALBPP)
#elif UNCOMP>0
#define HandleZRLE CONCAT3E(HandleZRLE,REALBPP,Down)
#define HandleZRLETile CONCAT3E(HandleZRLETile,REALBPP,Down)
#else
#define HandleZRLE CONCAT3E(HandleZRLE,REALBPP,Up)
#define HandleZRLETile CONCAT3E(HandleZRLETile,REALBPP,Up)
#endif
#undef CARDBPP
#undef CARDREALBPP
#define CARDBPP CONCAT2E(CARD, BPP)
#define CARDREALBPP CONCAT2E(CARD,REALBPP)

#define FillRectangle(x, y, w, h, color)  \
	{ \
		XGCValues _gcv; \
		_gcv.foreground = color; \
		if (!appData.useBackingstore) { \
			FillScreen(x, y, w, h, _gcv.foreground); \
		} else { \
			XChangeGC(dpy, gc, GCForeground, &_gcv); \
			XFillRectangle(dpy, desktopWin, gc, x, y, w, h); \
		} \
	}

static int HandleZRLETile(
	unsigned char* buffer,size_t buffer_length,
	int x,int y,int w,int h);

static Bool
HandleZRLE (int rx, int ry, int rw, int rh)
{
	rfbZRLEHeader header;
	int remaining;
	int inflateResult;
	int toRead;
	int min_buffer_size = rw * rh * (REALBPP / 8) * 2;

	/* First make sure we have a large enough raw buffer to hold the
	 * decompressed data.  In practice, with a fixed REALBPP, fixed frame
	 * buffer size and the first update containing the entire frame
	 * buffer, this buffer allocation should only happen once, on the
	 * first update.
	 */
	if ( raw_buffer_size < min_buffer_size) {

		if ( raw_buffer != NULL ) {

			free( raw_buffer );

		}

		raw_buffer_size = min_buffer_size;
		raw_buffer = (char*) malloc( raw_buffer_size );

	}

	if (!ReadFromRFBServer((char *)&header, sz_rfbZRLEHeader))
		return False;

	remaining = Swap32IfLE(header.length);

	/* Need to initialize the decompressor state. */
	decompStream.next_in   = ( Bytef * )buffer;
	decompStream.avail_in  = 0;
	decompStream.next_out  = ( Bytef * )raw_buffer;
	decompStream.avail_out = raw_buffer_size;
	decompStream.data_type = Z_BINARY;

	/* Initialize the decompression stream structures on the first invocation. */
	if ( decompStreamInited == False ) {

		inflateResult = inflateInit( &decompStream );

		if ( inflateResult != Z_OK ) {
			fprintf(stderr, 
					"inflateInit returned error: %d, msg: %s\n",
					inflateResult,
					decompStream.msg);
			return False;
		}

		decompStreamInited = True;

	}

	inflateResult = Z_OK;

	/* Process buffer full of data until no more to process, or
	 * some type of inflater error, or Z_STREAM_END.
	 */
	while (( remaining > 0 ) &&
			( inflateResult == Z_OK )) {

		if ( remaining > BUFFER_SIZE ) {
			toRead = BUFFER_SIZE;
		}
		else {
			toRead = remaining;
		}

		/* Fill the buffer, obtaining data from the server. */
		if (!ReadFromRFBServer(buffer,toRead))
			return False;

		decompStream.next_in  = ( Bytef * )buffer;
		decompStream.avail_in = toRead;

		/* Need to uncompress buffer full. */
		inflateResult = inflate( &decompStream, Z_SYNC_FLUSH );

		/* We never supply a dictionary for compression. */
		if ( inflateResult == Z_NEED_DICT ) {
			fprintf(stderr, "zlib inflate needs a dictionary!\n");
			return False;
		}
		if ( inflateResult < 0 ) {
			fprintf(stderr, 
					"zlib inflate returned error: %d, msg: %s\n",
					inflateResult,
					decompStream.msg);
			return False;
		}

		/* Result buffer allocated to be at least large enough.  We should
		 * never run out of space!
		 */
		if (( decompStream.avail_in > 0 ) &&
				( decompStream.avail_out <= 0 )) {
			fprintf(stderr, "zlib inflate ran out of space!\n");
			return False;
		}

		remaining -= toRead;

	} /* while ( remaining > 0 ) */

	if ( inflateResult == Z_OK ) {
		void* buf=raw_buffer;
		int i,j;

		remaining = raw_buffer_size-decompStream.avail_out;

		for(j=0; j<rh; j+=rfbZRLETileHeight)
			for(i=0; i<rw; i+=rfbZRLETileWidth) {
				int subWidth=(i+rfbZRLETileWidth>rw)?rw-i:rfbZRLETileWidth;
				int subHeight=(j+rfbZRLETileHeight>rh)?rh-j:rfbZRLETileHeight;
				int result=HandleZRLETile(buf,remaining,rx+i,ry+j,subWidth,subHeight);

				if(result<0) {
					fprintf(stderr, "ZRLE decoding failed (%d)\n",result);
return True;
					return False;
				}

				buf+=result;
				remaining-=result;
			}
	}
	else {

		fprintf(stderr, 
				"zlib inflate returned error: %d, msg: %s\n",
				inflateResult,
				decompStream.msg);
		return False;

	}

	return True;
}

#if REALBPP!=BPP && defined(UNCOMP) && UNCOMP!=0
#if UNCOMP>0
#define UncompressCPixel(pointer) ((*(CARDBPP*)pointer)>>UNCOMP)
#else
#define UncompressCPixel(pointer) ((*(CARDBPP*)pointer)<<(-(UNCOMP)))
#endif
#else
#define UncompressCPixel(pointer) (*(CARDBPP*)pointer)
#endif

static int HandleZRLETile(
		unsigned char* buffer,size_t buffer_length,
		int x,int y,int w,int h) {
	unsigned char* buffer_copy = buffer;
	unsigned char* buffer_end = buffer+buffer_length;
	unsigned char type;

	if(buffer_length<1)
		return -2;

	if (frameBufferLen < w * h * REALBPP/8) {
		if(frameBuffer) {
			free(frameBuffer);
		}
		frameBufferLen = w * h * REALBPP/8 * 2;
		frameBuffer = (unsigned char *) malloc(frameBufferLen);
	}

	type = *buffer;
	buffer++;
	switch(type) {
		case 0: /* raw */
		{
#if REALBPP!=BPP
			int m0 = 0, i,j;

			if(1+w*h*REALBPP/8>buffer_length) {
				fprintf(stderr, "expected %d bytes, got only %d (%dx%d)\n",1+w*h*REALBPP/8,buffer_length,w,h);
				return -3;
			}

			for(j=y*si.framebufferWidth; j<(y+h)*si.framebufferWidth; j+=si.framebufferWidth) {
				for(i=x; i<x+w; i++,buffer+=REALBPP/8) {
#  if 0
					((CARDBPP*)frameBuffer)[j+i] = UncompressCPixel(buffer);
					/* alt */
					CARDBPP color = UncompressCPixel(buffer);
					CopyDataToScreen((char *)&color, i, j/si.framebufferWidth, 1, 1);
#  else
					((CARDBPP*)frameBuffer)[m0++] = UncompressCPixel(buffer);
#  endif
				}
			}
			CopyDataToScreen((char *)frameBuffer, x, y, w, h);
if (0) fprintf(stderr, "cha1: %dx%d+%d+%d\n", w, h, x, y);

#else
#  if 0
			CopyRectangle(buffer, x, y, w, h);
#  else
			CopyDataToScreen((char *)buffer, x, y, w, h);
#  endif
			buffer+=w*h*REALBPP/8;
#endif
			break;
		}
		case 1: /* solid */
		{
			CARDBPP color = UncompressCPixel(buffer);

			if(1+REALBPP/8>buffer_length)
				return -4;
				
			if ((BPP == 8 && appData.useBGR233) || (BPP == 16 && appData.useBGR565)) {
				int m0;
				for (m0=0; m0 < w*h; m0++) {
					((CARDBPP*)frameBuffer)[m0] = color;
				}
				CopyDataToScreen((char *)frameBuffer, x, y, w, h);
			} else {
				FillRectangle(x, y, w, h, color);
			}
if (0) fprintf(stderr, "cha2: %dx%d+%d+%d\n", w, h, x, y);

			buffer+=REALBPP/8;

			break;
		}
		case 2 ... 127: /* packed Palette */
		{
			CARDBPP palette[16];
			int m0, i,j,shift,
				bpp=(type>4?(type>16?8:4):(type>2?2:1)),
				mask=(1<<bpp)-1,
				divider=(8/bpp);

			if(1+type*REALBPP/8+((w+divider-1)/divider)*h>buffer_length)
				return -5;

			/* read palette */
			for(i=0; i<type; i++,buffer+=REALBPP/8)
				palette[i] = UncompressCPixel(buffer);

			m0 = 0;
			/* read palettized pixels */
			for(j=y*si.framebufferWidth; j<(y+h)*si.framebufferWidth; j+=si.framebufferWidth) {
				for(i=x,shift=8-bpp; i<x+w; i++) {
#  if 0
					((CARDBPP*)frameBuffer)[j+i] = palette[((*buffer)>>shift)&mask];
					/* alt */
					CARDBPP color = palette[((*buffer)>>shift)&mask];
					CopyDataToScreen((char *)&color, i, j/si.framebufferWidth, 1, 1);
#  else
					((CARDBPP*)frameBuffer)[m0++] = palette[((*buffer)>>shift)&mask];
#  endif
					shift-=bpp;
					if(shift<0) {
						shift=8-bpp;
						buffer++;
					}
				}
				if(shift<8-bpp)
					buffer++;
			}
			CopyDataToScreen((char *)frameBuffer, x, y, w, h);
if (0) fprintf(stderr, "cha3: %dx%d+%d+%d\n", w, h, x, y);

			break;
		}
		/* case 17 ... 127: not used, but valid */
		case 128: /* plain RLE */
		{
			int m0=0, i=0,j=0;
			while(j<h) {
				int color,length;
				/* read color */
				if(buffer+REALBPP/8+1>buffer_end)
					return -7;
				color = UncompressCPixel(buffer);
				buffer+=REALBPP/8;
				/* read run length */
				length=1;
				while(*buffer==0xff) {
					if(buffer+1>=buffer_end)
						return -8;
					length+=*buffer;
					buffer++;
				}
				length+=*buffer;
				buffer++;
				while(j<h && length>0) {
#  if 0
					((CARDBPP*)frameBuffer)[(y+j)*si.framebufferWidth+x+i] = color;
					/* alt */
					CopyDataToScreen((char *)&color, x+i, y+j, 1, 1);
#  else
					((CARDBPP*)frameBuffer)[m0++] = color;
#  endif
					length--;
					i++;
					if(i>=w) {
						i=0;
						j++;
					}
				}
				if(length>0)
					fprintf(stderr, "Warning: possible ZRLE corruption\n");
			}
			CopyDataToScreen((char *)frameBuffer, x, y, w, h);
if (0) fprintf(stderr, "cha4: %dx%d+%d+%d\n", w, h, x, y);

			break;
		}
		case 129: /* unused */
		{
			return -8;
		}
		case 130 ... 255: /* palette RLE */
		{
			CARDBPP palette[128];
			int m0 = 0, i,j;

			if(2+(type-128)*REALBPP/8>buffer_length)
				return -9;

			/* read palette */
			for(i=0; i<type-128; i++,buffer+=REALBPP/8)
				palette[i] = UncompressCPixel(buffer);
			/* read palettized pixels */
			i=j=0;
			while(j<h) {
				int color,length;
				/* read color */
				if(buffer>=buffer_end)
					return -10;
				color = palette[(*buffer)&0x7f];
				length=1;
				if(*buffer&0x80) {
					if(buffer+1>=buffer_end)
						return -11;
					buffer++;
					/* read run length */
					while(*buffer==0xff) {
						if(buffer+1>=buffer_end)
							return -8;
						length+=*buffer;
						buffer++;
					}
					length+=*buffer;
				}
				buffer++;
				while(j<h && length>0) {
#  if 0
					((CARDBPP*)frameBuffer)[(y+j)*si.framebufferWidth+x+i] = color;
					/* alt */
					CopyDataToScreen((char *)&color, x+i, y+j, 1, 1);
#  else
					((CARDBPP*)frameBuffer)[m0++] = color;
#  endif
					length--;
					i++;
					if(i>=w) {
						i=0;
						j++;
					}
				}
				if(length>0)
					fprintf(stderr, "Warning: possible ZRLE corruption\n");
			}
			CopyDataToScreen((char *)frameBuffer, x, y, w, h);
if (0) fprintf(stderr, "cha5: %dx%d+%d+%d\n", w, h, x, y);

			break;
		}
	}

	return buffer-buffer_copy;	
}

#undef CARDBPP
#undef CARDREALBPP
#undef HandleZRLE
#undef HandleZRLETile
#undef UncompressCPixel
#undef REALBPP

#undef UNCOMP

#undef FillRectangle
